// ==========================================================================
// $Id: vector3d.h,v 1.1 2013/11/06 17:22:56 jlang Exp $
// CSI2372 example Code for lecture 13
// ==========================================================================
// (C)opyright:
//
//   Jochen Lang
//   SITE, University of Ottawa
//   800 King Edward Ave.
//   Ottawa, On., K1N 6N5
//   Canada. 
//   http://www.site.uottawa.ca
// 
// Creator: jlang (Jochen Lang)
// Email:   jlang@site.uottawa.ca
// ==========================================================================
// $Log: vector3d.h,v $
// Revision 1.1  2013/11/06 17:22:56  jlang
// Created templated vector example
//
// Revision 1.2  2006/11/20 21:34:07  jlang
// Added friend/inline/static demo
//
//
// ==========================================================================
#ifndef VECTOR_3D_
#define VECTOR_3D_

#include <iostream>
#include <cmath>

using std::ostream;
using std::endl;


template <class T, const int NUM> class vec {
  T d_comp[NUM];
 public:
  inline vec( T* _in );
  inline vec();
  inline T& operator()(int);
  inline T operator()(int) const;
  operator T* ();
  operator const T* ();
};


template <class T, const int NUM>
  vec<T,NUM>::vec(T* _in) {
  for (int i=0;i<NUM; ++i) {
    d_comp[i] = _in[i];
  }
}


template <class T, const int NUM>
  vec<T,NUM>::vec() {
  for (int i=0;i<NUM; ++i) {
    d_comp[i] = 0;
  }
}


template <class T, const int NUM>
  T& vec<T,NUM>::operator()(int i) { 
  return d_comp[i]; 
}

template <class T, const int NUM>
  T vec<T,NUM>::operator()(int i) const { 
  return d_comp[i]; 
}


template <class T, const int NUM>
vec<T,NUM>::operator T* () { 
  return d_comp; 
} 

template <class T, const int NUM>
vec<T,NUM>::operator const T* () { 
  return d_comp; 
} 

template <class T>
std::ostream& operator<<( std::ostream& _out, const vec<T,3>& _pt );
template <class T>
std::istream& operator>>( std::istream& _in, vec<T,3>& _pt );


template <class T> class vec<T,3> {
  T d_comp[3];
  static const double ZERO;
 public:
  // constructors
  inline vec( const T _in[3] );
  inline vec( const T _x, const T _y, const T _z );
  inline vec();
  // accessors
  inline T& x() { return d_comp[0]; }
  inline T& y() { return d_comp[1]; }
  inline T& z() { return d_comp[2]; }
  inline T x() const { return d_comp[0]; }
  inline T y() const { return d_comp[1]; }
  inline T z() const { return d_comp[2]; }
  inline T& operator()(int i) { return d_comp[i]; }
  inline T operator()(int i) const { return d_comp[i]; }
  // arithmetic operators
  inline vec<T,3>& operator+=( const vec<T,3>& _pt );
  inline vec<T,3>& operator-=( const vec<T,3>& _pt );
  inline vec<T,3>& operator*=( const T& _scalar );
  inline vec<T,3> cross( const vec<T,3>& _b );
  inline T lengthSqr();
  inline double length();
  inline bool operator!();
  // conversion operators
  inline operator T* ();
  inline operator const T* ();
  // conversion operator to bool -- enables if ( vec )
  inline operator bool(); 

  // io functions
  friend std::ostream& operator<< <T> ( std::ostream& _out, const vec<T,3>& _pt );
  friend std::istream& operator>> <T> ( std::istream& _in, vec<T,3>& _pt );
};



template <class T>
const double vec<T,3>::ZERO = 1E-99;


//********************************
// Constructors
//********************************
template <class T>
  vec<T,3>::vec(const T _in[3]) {
  d_comp[0] = _in[0];
  d_comp[1] = _in[1];
  d_comp[2] = _in[2];
 }

template <class T>
  vec<T,3>::vec( const T _x,const T _y, const T _z ) {
  d_comp[0] = _x;
  d_comp[1] = _y;
  d_comp[2] = _z;
 }

template <class T>
  vec<T,3>::vec() {
  for (int i=0;i<3; ++i) {
    d_comp[i] = 0;
  }
 }



//***********************************
// Arithmetic operators and functions
//***********************************
template <class T>
vec<T,3>& vec<T,3>::operator+=( const vec<T,3>& _pt ) {
  for ( int i=0; i<3; ++i ) {
    d_comp[i] += _pt.d_comp[i];
  }
  return *this;
 }

template <class T>
vec<T,3>& vec<T,3>::operator-=( const vec<T,3>& _pt ) {
  for ( int i=0; i<3; ++i ) {
    d_comp[i] -= _pt.d_comp[i];
  }
  return *this;
 }

template <class T>
vec<T,3>& vec<T,3>::operator*=( const T& scalar  ) {
  for ( int i=0; i<3; ++i ) {
    d_comp[i] *= scalar;
  }
  return *this;
 }

template <class T>
vec<T,3>  vec<T,3>::cross( const vec<T,3>& _b ) {
  vec<T,3> res;
  res.x() = y()*_b.z() - z()*_b.y();
  res.y() = - x()*_b.z() + z()*_b.x();
  res.z() = x()*_b.y() - y()*_b.x();
  return res;
 }

template <class T>
T vec<T,3>::lengthSqr( ) {
  T res = 0;
  for ( int i=1; i<3; ++i ) {
    res += d_comp[i] * d_comp[i];
  }
  return res;
}

template <class T>
double vec<T,3>::length( ) {
  return sqrt(static_cast<double>(lengthSqr()));
}

template <class T>
bool vec<T,3>::operator!() {
  for ( int i=0; i<3; ++i ) {
    if ( fabs(d_comp[i]) > ZERO ) return false;
  }
  return true;
}

//************************************
// Conversion operators
//************************************
template <class T>
vec<T,3>::operator bool() { 
  return !(!*this); 
};


template <class T>
vec<T,3>::operator T* () { 
  return d_comp; 
} 

template <class T>
vec<T,3>::operator const T* () { 
  return d_comp; 
} 


//************************************
// Insertion and extraction operators
//************************************
template <class T>
std::ostream& operator<<( std::ostream& _out, const vec<T,3>& _pt ) {
  _out << _pt.d_comp[0] << " " << _pt.d_comp[1] << " " 
       << _pt.d_comp[2] << " ";
  return _out;
}
  

template <class T>
std::istream& operator>>( std::istream& _in, vec<T,3>& _pt ) {
  _in >> _pt.d_comp[0] >> _pt.d_comp[1] >> _pt.d_comp[2];
  if ( !_in ) _pt=vec<T,3>();
  return _in;
}

typedef vec<double,3> Vector3D; 

#endif
